/* Copyright (C) 2007 Damien Good. This module is free software; you
can redistribute it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version. This
module is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details. You should have received a copy of the GNU Lesser General Public
License along with this module; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA */

#include "stdafx.h"

#include "gme/Music_Emu.h"

#include "Wave_Writer.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

char strdata[513];
int commandlinewidth = 80;

void handle_error( const char* str );

void usage(char *name)
{
	printf("Usage: %s [options] [music files]\n\n",name);
	printf("[options]\n");
	printf("-V [seconds]     Additional time to check (default 180 seconds)\n");
	printf("-L [Loops]       Loops to calculate for Time: (default 2 loops)\n");
	printf("-clw [number]    Set command line width. (default is 80)\n");

	printf("-?               This screen\n\n");
	printf("This program Uses Blargg's Game_Music_Emu library\n");
	printf("Available at http://slack.net/~ant/\n\n");
	printf("File formats currently supported\n");
	printf("Super Nintendo, *.spc\n");
}

void printstring( char *data )
{
	int i;
	if(strlen(data)>(commandlinewidth-1))
	{
		for(i=0;i<(commandlinewidth-1);i++)
			printf("%c",data[strlen(data)-(commandlinewidth-1)+i]);
	}
	else
		printf("%s",data);
}

int main(int argc, char *argv[])
{
	const long size = 32768; // can be any multiple of 2
			short buf [size];
	long sample_rate = 44100; // number of samples per second
	int track = 0; // index of track to play (0 = first)
	int PlayTime;
	int BPplaytime;
	int BPlooptime;
	int BPintrotime;
	int BPfinaltime;
	int OrigLoopCount = 2; //2 loops by default
	int UsedLoopCount;
	int minutes,seconds,mseconds;
	int i,j,k;
	int bytescountequal,verify_loops=485;
	int bytescountequal2;
	
	int maskoutput=0;
	unsigned char maskbytes[32];

	int debugmap=0;

	FILE *f;

	track_info_t info;

	int prev_count[3][2];

	int fileindex;

	printf("Game Music Emulator Timer\n");
	printf("Copyright 2007 CaitSith2\n\n");
	

	if(argc<2)
	{
		printf("Usage: %s [options] [music files]\n",argv[0]);
		printf("For extended usage info, %s -?\n",argv[0]);
		return 0;
	}

	for(i=1;i<argc;i++)
	{
		if(!stricmp(argv[i],"-V"))
		{
			verify_loops = (atoi(argv[++i]) * sample_rate * 2) / size;
			continue;
		}
		if(!stricmp(argv[i],"-L"))
		{
			OrigLoopCount = atoi(argv[++i]);
			if (OrigLoopCount < 1) OrigLoopCount = 1;
			//Overflows at maximum time(99:59.999) using 357,
			//by my calculations, but just being sane.
			if (OrigLoopCount > 300) OrigLoopCount = 300;
			continue;
		}
		if(!stricmp(argv[i],"-?"))
		{
			usage(argv[0]);
			return 0;
		}

		if(!stricmp(argv[i],"-clw"))
		{
			commandlinewidth = atoi(argv[++i]);
			if(commandlinewidth>512)
				commandlinewidth=512;
			continue;
		}

		if(!stricmp(argv[i],"-mask")) //TODO: Document feature
		{
			maskoutput=1;
			continue;
		}

		if(!stricmp(argv[i],"-debugmap")) //Don't document
		{
			debugmap=atoi(argv[++i]);
			continue;
		}

		

		fileindex=i;
		break;
	}

	

	for(;fileindex<argc;fileindex++)
	{
		for(i=0;i<6;i++)
			prev_count[i/2][i%2]=0;
	
		// Determine file type
		gme_type_t file_type;
		handle_error( gme_identify_file( argv[fileindex], &file_type ) );
		if ( !file_type )
			handle_error( "Unsupported music type" );
		
		// Create emulator and set sample rate
		Music_Emu* emu = file_type->new_emu();
		if ( !emu )
			handle_error( "Out of memory" );
		handle_error( emu->set_sample_rate( sample_rate ) );
		
		// Load music file into emulator
		handle_error( emu->load_file( argv[fileindex] ) );
		
		// Start track
		handle_error( emu->start_track( track ) );

		if(emu->read_count(0)==-1)
		{
			delete emu;
			printf("Timing not supported for %s\n",argv[fileindex]);
			continue;
		}
		
		// Begin writing to wave file
		//Wave_Writer wave( sample_rate, "out.wav" );
		//wave.enable_stereo();
		
		// Record 10 seconds of track

		emu->track_info(&info);

		if(!stricmp("Super Nintendo",info.system))
			if(maskoutput)
			{
				emu->read_count((3<<24)+1);
			}

	//	sprintf(strdata,"%s.debug",argv[fileindex]);
	//	f=fopen(strdata,"wb");

		while ( (PlayTime = emu->tell()) < 99 * 60 * 1000L )
		{
/*
			if(f!=NULL)
			{
				fwrite(&PlayTime,1,4,f);
				for(i=0;i<3;i++)
				{
					j=emu->read_count(i);
					fwrite(&j,1,4,f);
				}
			}*/

			
				
			
			if(1)
			{
				if((emu->read_count(0)>=0)&&(emu->read_count(0)<=0x10000)) //THe counts sometimes bugger up
				{														   //Just play one more tick silently if it does.
					j=0;
						if(prev_count[0][0] == emu->read_count(0)) j++;
						prev_count[0][0] = emu->read_count(0);
						if(prev_count[1][0] == emu->read_count(1) + emu->read_count(2)) j+=2;
						prev_count[1][0] = emu->read_count(1) + emu->read_count(2);
					if(j==3)
						bytescountequal++;
					else
					{
						BPplaytime = PlayTime;
						bytescountequal=0;
					}
					for(i=0,j=0;i<3;i++)
					{
						if(prev_count[i][1] == emu->read_count(i)) j++;
						prev_count[i][1] = emu->read_count(i);
					}
					if(j==3)
						bytescountequal2++;
					else
					{
						BPlooptime = PlayTime;
						bytescountequal2=0;
					}
				}

				//Ugetab
				//Just making use of a variable, to keep coding consistent
				BPintrotime = (BPplaytime - (BPlooptime - BPplaytime));

				minutes = BPlooptime / (60*1000);
				seconds = (BPlooptime % (60*1000)) / 1000;
				mseconds = BPlooptime % 1000;
				for(i=0;i<(commandlinewidth+2);i++) printf("\b");
				sprintf(strdata,"%s: ",argv[fileindex]);
				sprintf(strdata,"%sTime: %d:%.2d.%.3d, ",strdata,minutes,seconds,mseconds);
						minutes = (BPintrotime) / (60*1000);
						seconds = ((BPintrotime) % (60*1000)) / 1000;
						mseconds = (BPintrotime) % 1000;
						if((minutes >= 0) && (seconds >= 0))
						{
							if(mseconds<0) mseconds = 0;
							sprintf(strdata,"%sIntro: %d:%.2d.%.3d, ",strdata,minutes,seconds,mseconds);
							minutes = ((BPlooptime - BPplaytime)) / (60*1000);
							seconds = (((BPlooptime - BPplaytime)) % (60*1000)) / 1000;
							mseconds = ((BPlooptime - BPplaytime)) % 1000;
							sprintf(strdata,"%sLoop: %d:%.2d.%.3d, ",strdata,minutes,seconds,mseconds);
						}
						else
						{
							sprintf(strdata,"%sIntro/Loop info not available, ",strdata);
						}
				seconds = (((verify_loops - bytescountequal2) * (size/2)) / sample_rate);
				minutes = seconds / 60;
				seconds %= 60;
				sprintf(strdata,"%s%d:%.2d  ",strdata,minutes,seconds);
				printstring(strdata);
				if((bytescountequal2==verify_loops)||emu->TrackEnded)
				{
					//Playback ended, more intensive calculations can now be done
					for(i=0;i<(commandlinewidth+2);i++) printf("\b");
					for(i=0;i<(commandlinewidth-1);i++) printf(" ");
					for(i=0;i<(commandlinewidth+2);i++) printf("\b");
					sprintf(strdata,"%s: ",argv[fileindex]);

					if(emu->TrackEnded)
					{
						//Ugetab: Because of how winamp handles the timing, all
						//'one shot' ending times should be rounded up to the
						//next full digit. If you don't, one shots get truncated
						if ((BPlooptime % 1000) > 0)
						{
							BPlooptime += 1000;
							BPlooptime -= (BPlooptime % 1000);
						}

						minutes = ((BPlooptime)) / (60*1000);
						seconds = (((BPlooptime)) % (60*1000)) / 1000;
						mseconds = ((BPlooptime)) % 1000;

						sprintf(strdata,"%sTime: %d:%.2d.%.3d, Song is one shot",strdata,minutes,seconds,mseconds);
					}
					else
					{
						//Ugetab:
						//Subtract Loop time from Intro time until Intro is less than Loop,
						//but only if the intro will be less than 30, and >= 0 seconds,
						//otherwise make the user check the loop point themselves.
						//The 30 seconds part is because if the Intro is longer than that
						//after the loop amount is subtracted, the program is likely wrong,
						//and if it's longer validly, then the program may not handle it well
						BPintrotime = (BPplaytime - (BPlooptime - BPplaytime));
						
						//If the loop is bigger than the intro to start with,
						//check if it's within .010 seconds of being the same
						if (((BPlooptime - BPplaytime) > BPintrotime) && (((BPlooptime - BPplaytime) - BPintrotime) <= 10))
						{
							//If it's within .010 seconds of being the same,
							//make the intro equal to the loop time.
							BPintrotime = (BPlooptime - BPplaytime);
						}

						while ((BPintrotime >= (BPlooptime - BPplaytime)) && (((((BPintrotime) - (BPlooptime - BPplaytime)) % (60*1000)) / 1000) < 30) && (((((BPintrotime) - (BPlooptime - BPplaytime)) % (60*1000)) / 1000) >= 0)) BPintrotime -= (BPlooptime - BPplaytime);
						
						//Check if it's within 5 seconds of being the same
						//You can change the number of loops applied here. Default is * 2.
						//1 loop minimum.
						//OrigLoopCount = 2;
						UsedLoopCount = OrigLoopCount;

						BPfinaltime = BPintrotime + ((BPlooptime - BPplaytime) * UsedLoopCount);
						
						//Only a very short track could have a difference of 5
						//seconds in either direction, and have this break it.
						//Just have people check relatively short tracks.
						//If Loop Time is >= 5 seconds:
						if ((BPlooptime - BPplaytime) >= 5000)
						{
						//Up to 10 seconds shorter or longer, reduce number of loops used by 1, keep intro as is:
						if ((((BPlooptime - BPplaytime) > BPintrotime) && (((BPlooptime - BPplaytime) - BPintrotime) <= 5000)) || ((BPintrotime > (BPlooptime - BPplaytime)) && ((BPintrotime - (BPlooptime - BPplaytime)) <= 5000)))
						{
						//Should be multiplied by 1 less than the original.
						UsedLoopCount -= 1;
						BPfinaltime = BPintrotime + ((BPlooptime - BPplaytime) * UsedLoopCount);
						}
						}

						minutes = ((BPfinaltime)) / (60*1000);
						seconds = (((BPfinaltime)) % (60*1000)) / 1000;
						mseconds = ((BPfinaltime)) % 1000;
						sprintf(strdata,"%sTime: %d:%.2d.%.3d, ",strdata,minutes,seconds,mseconds);
						minutes = (BPintrotime) / (60*1000);
						seconds = ((BPintrotime) % (60*1000)) / 1000;
						mseconds = (BPintrotime) % 1000;
						if((minutes >= 0) && (seconds >= 0))
						{
							if(mseconds<0) mseconds = 0;
							sprintf(strdata,"%sIntro: %d:%.2d.%.3d, ",strdata,minutes,seconds,mseconds);
							minutes = ((BPlooptime - BPplaytime)) / (60*1000);
							seconds = (((BPlooptime - BPplaytime)) % (60*1000)) / 1000;
							mseconds = ((BPlooptime - BPplaytime)) % 1000;
							sprintf(strdata,"%sLoop: %d:%.2d.%.3d(%d)",strdata,minutes,seconds,mseconds,UsedLoopCount);
						}
						else
						{
							sprintf(strdata,"%sIntro/Loop info not available",strdata);
						}
					}

					//sprintf(strdata,"%d, %d",BPlooptime,info.length);

					if(info.length != -1)
					{
						if((BPlooptime + (20 * 1000)) < info.length)
							sprintf(strdata,"%s, Check",strdata);
						else if ((info.length + (20 * 1000)) < BPlooptime)
							sprintf(strdata,"%s, Check",strdata);
					}
					else
					{
						sprintf(strdata,"%s, No Timer present",strdata);
					}
					sprintf(strdata,"%s\n",strdata);

					printstring(strdata);
					break;

				}
			}
			
			// Fill buffer
			handle_error( emu->play( size, buf ) );
			
			// Write samples to wave file
			//wave.write( buf, size );
		}

		//fclose(f);
		
		if(!stricmp("Super Nintendo",info.system))
			if(maskoutput)
			{
				sprintf(strdata,"%s.mask",argv[fileindex]);
				for(i=0;i<32;i++)
				{
					maskbytes[i] = 0;
					for(j=0;j<8;j++)
					{
						for(k=0;k<256;k++)
							if(((emu->read_count((1<<24)+((i*2048)+(j*256)+k)))>0)&&((emu->read_count((1<<24)+((i*2048)+(j*256)+k)))<255))
								break;
						if(k!=256)
							maskbytes[i] |= (1 << (7 - j));
					}
				}
			f=fopen(strdata,"wb");
			if(f!=NULL)
			{
				fwrite(maskbytes,1,32,f);
				fclose(f);
			}
		}

		if(debugmap)
		{
			sprintf(strdata,"%s.%d.map",argv[fileindex],debugmap);
			f=fopen(strdata,"wb");
			if(f!=NULL)
			{
				for(i=0;i<0x10000;i++)
					fputc(emu->read_count((1<<24)+i),f);
				fclose(f);
			}
		}

		/*sprintf(strdata,"%s.map",argv[fileindex]);
		f=fopen(strdata,"wb");
		for(i=0;i<0x10000;i++)
			fputc(emu->read_count((1<<24)+i),f);
		fclose(f);*/
		
		// Cleanup
		delete emu;
	}
	
	return 0;
}

void handle_error( const char* str )
{
	if ( str )
	{
		printf( "Error: %s\n", str ); getchar();
		exit( EXIT_FAILURE );
	}
}

